
"""User LDAP product

This product provides support for User Folder-like objects that
store data in LDAP Directories.

"""

__version__='$Revision: 1.2 $'[11:-2]

import Globals, App.Undo, socket, regex
from Products.ZLDAPMethods import LM 

from Globals import PersistentMapping, Persistent, HTMLFile, MessageDialog
from string import join, strip, split, lower, upper
from App.Management import Navigation, Tabs
from Acquisition import Implicit
from OFS.SimpleItem import Item
from OFS.Document import Document
from base64 import decodestring, encodestring
from urllib import quote, unquote
from ImageFile import ImageFile
from AccessControl.Role import RoleManager
from AccessControl.User import BasicUser, BasicUserFolder, domainSpecMatch
import sha, binascii

### Crypt may or not be compiled in. 
try:
    from crypt import crypt
except ImportError:
    crypt=None			#If crypt is not compiled in, we'll check later

UserLDAPError = 'UserLDAP Error'
ListType=type([])


class User(BasicUser):
    
    icon='misc_/UserLDAP/User_icon'

    def __init__(self, name, password, roles=['Anonymous'], domains=[]):
	self.name   = name
        self.__     = password
        self.roles  = roles
        self.domains= domains

    def getUserName(self):
        return self.name

    def _getPassword(self):
        return self.__

    def getRoles(self):
        return self.roles
    
    def getDomains(self):
        return self.domains

    def authenticate(self, password, request,):
	
	passwrd=self._getPassword()
	result = 0
	if upper(passwrd[:5]) == '{SHA}':
	    password = binascii.b2a_base64(sha.new(password).digest())[:-1]
	    result = passwrd[5:] == password
	elif upper(passwrd[:7]) == '{CRYPT}' and crypt is not None:
	    #if crypt is None, it's not compiled in and everything will fail
	    password = crypt(password, passwrd[7:9])
	    result = passwrd[7:] == password
	else:
	    result = passwrd == password
        
	domains=self.getDomains()
        if domains:
            return result and domainSpecMatch(domains, request)
        return result



class UserLDAP(BasicUserFolder):
    """ """

    meta_type='User LDAP'
    id       ='acl_users'
    title    ='User LDAP'
    icon     ='misc_/UserLDAP/UserLDAP_icon'

    isPrincipiaFolderish=1
    isAUserFolder=1

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Properties', 'action':'manage_properties'},
    {'label':'Security',   'action':'manage_access'},
    {'label':'Undo',       'action':'manage_UndoForm'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_copyright', 'manage_tabs',
      'manage_properties', 'manage_UndoForm']),
    ('Undo changes',       ['manage_undo_transactions']),
    ('Change permissions', ['manage_access']),
    ('Manage users',       ['manage_users']),
    ('Change User LDAPs',     ['manage_edit']),
    )

    manage_properties=HTMLFile('properties', globals())
    _editUser=HTMLFile('editUser', globals())
    _mainUser=HTMLFile('mainUser', globals())
    manage=manage_main=_mainUser
    
    def __init__(self, conn_id, scope, basedn, 
		 user_name_attr, password_attr,
		 roles_attr, domains_attr,
		 cookie_mode=0):
        self.conn_id=conn_id
	self.scope = scope

	self.user_name_attr = user_name_attr
	self.password_attr = password_attr
	self.roles_attr = roles_attr
	self.domains_attr = domains_attr

        self.cookie_mode=cookie_mode
        
        self.docLogin =Document(_docLogin, __name__='docLogin')
        self.docLogout=Document(_docLogout, __name__='docLogout')
        self.docLogin.__roles__=None
        self.docLogout.__roles__=None

        
        self.ldapListQuery=LM.LDAPFilter('ldapListQuery', '',
                                         conn_id, scope, basedn,
                                         '', 'uid=*')


        self.ldapUserQuery=LM.LDAPFilter('ldapUserQuery', '',
                                         conn_id, scope, basedn,
                                         'name', 'uid=<!--#var name-->')

    def getUserNames(self):
        """Returns a list of user names or [] if no users exist"""
        data=[]
        try:    items=self.ldapListQuery()
	except: return data
        for ob in items:
	    try: data.append(str(getattr(ob, self.user_name_attr)))
	    except: pass
        data.sort()
        return data

    def getUsers(self):
        """Return a list of user objects or [] if no users exist"""
        data=[]
        try:    items=self.ldapListQuery()
        except: return data
        for ob in items:
	    try:
		n = str(getattr(ob, self.user_name_attr))
		p = str(getattr(ob, self.password_attr))
		r = ob.Get(self.roles_attr)
		d = ob.Get(self.domains_attr)
	    except KeyError, absent:
		if strip(join(absent.args)) == 'roles':
		    r = ['Anonymous']
		    continue
		elif strip(join(absent.args)) == 'domains':
		    d = []
		else:
		    raise NameError, absent
		
            user=User(n, p, r, d)
            data.append(user)
        return data

    def getUser(self, name):
        """Return the named user object or None if no such user exists"""

        try:    ob=self.ldapUserQuery(name=name)
        except: return None
        if not ob:
            return None
        ob=ob[0]
	n = str(getattr(ob, self.user_name_attr))
	p = str(getattr(ob, self.password_attr))
	try:
	    r = ob.Get(self.roles_attr)
	except KeyError, absent:
	    if strip(join(absent.args)) == self.roles_attr:
		r = ['Anonymous']
	    else:
		raise NameError, absent
	try:
	    d = getattr(ob, self.domains_attr)
	except KeyError, absent:
	    if strip(join(absent.args)) == self.domains_attr:
		d = []
	    else:
		raise NameError, absent
	    
        return User(n, p, r, d)

    def _doAddUser(self, name, password, roles, domains):
        raise UserLDAPError, 'Cannot Add New Users'

    def _doChangeUser(self, name, password, roles, domains):
        raise UserLDAPError, 'Cannot Change Users'

    def _doDelUsers(self, names):
        raise UserLDAPError, 'Cannot Delete Users'

# ----


    def old_std_validate(self,request,auth='',roles=None):
	parent=request['PARENTS'][0]

	# If no authorization, only a user with a
        # domain spec and no passwd or nobody can
        # match
	if not auth:
            for ob in self.getUsers():
                domains=ob.getDomains()
                if domains:
                    if ob.authenticate('', request):
                        if ob.allowed(parent, roles):
                            ob=ob.__of__(self)
                            return ob
            return None

	# Only do basic authentication
	if lower(auth[:6])!='basic ':
	    return None
	name,password=tuple(split(decodestring(split(auth)[-1]), ':'))

        user=self.getUser(name)
        if user is None:
            return None

	# Try to authenticate user
	if not user.authenticate(password, request):
	    return None

        # We need the user to be able to acquire!
        user=user.__of__(self)

	# Try to authorize user
	if user.allowed(parent, roles):
	    return user
	return None

    std_validate=BasicUserFolder.validate

    def cookie_validate(self,request,auth='',roles=None):
	parent=request['PARENTS'][0]
        req_has=request.has_key
        resp=request['RESPONSE']
        
        if req_has('__ac'):
            c=request['__ac']
            c=unquote(c)
            try: c=decodestring(c)
            except:
                resp.expireCookie('__ac', path='/')
                raise 'LoginRequired', self.docLogin(self, request)
            
            name,password=tuple(split(c, ':'))
            
            # Find user
            user=self.getUser(name)
            if user is not None:
                if user.authenticate(password, request):
                    user=user.__of__(self)
                    if user.allowed(parent, roles):
                        return user
            resp.expireCookie('__ac', path='/')
            raise 'LoginRequired', self.docLogin(self, request)

        if req_has('__ac_name') and req_has('__ac_password'):
            name=request['__ac_name']
            password=request['__ac_password']

            # Find user
            user=self.getUser(name)
            if user is not None:
                if user.authenticate(password, request):
                    user=user.__of__(self)
                    if user.allowed(parent, roles):
                        token='%s:%s' % (name, password)
                        token=encodestring(token)
                        token=quote(token)
                        resp.setCookie('__ac', token, path='/')
                        request['__ac']=token
                        try:
                            del request['__ac_name']
                            del request['__ac_password']
                        except:
                            pass
                        return user
            raise 'LoginRequired', self.docLogin(self, request)
        
        if auth: return None

        # Try domain matching
        for ob in self.getUsers():
            domains=ob.getDomains()
            if ob.domains:
                if ob.authenticate('', request):
                    if ob.allowed(parent, roles):
                        ob=ob.__of__(self)
                        return ob

        # Allow anonymous access to things under UserLDAPs,
        # if the object is public.
        nobody=self._nobody
        if nobody.allowed(parent, roles):
            ob=nobody.__of__(self)
            return ob

        raise 'LoginRequired', self.docLogin(self, request)


    def validate(self,request,auth='',roles=None):
        if self.cookie_mode:
            return self.cookie_validate(request, auth, roles)
        return self.std_validate(request, auth, roles)
    

    def __len__(self):
        return len(self.ldapListQuery())

    def enumQueries(self):
        list=[]
        type=LM.LDAP.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def enumDocuments(self):
        list=[]
        type=Document.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def manage_edit(self, connection_id, 
		    user_name_attr, password_attr,
		    roles_attr, domains_attr,
		    cookie_mode, REQUEST=None):
	"""Change properties"""
        if connection_id != self.conn_id:
            # Connection changed - update ldap methods!
            self.conn_id=connection_id
            for ob in self.enumQueries():
                title=ob.title
                args=ob.arguments_src
                src=ob.src
                ob.manage_edit(title, self.conn_id, args, src)
	
	self.user_name_attr = user_name_attr
	self.password_attr = password_attr
	self.roles_attr = roles_attr
	self.domains_attr = domains_attr

        self.cookie_mode=cookie_mode
        if REQUEST:
            return MessageDialog(
                   title  ='UserLDAP Changed',
                   message='UserLDAP properties have been updated',
                   action =REQUEST['URL1']+'/manage_properties',
                   target ='manage_main')

    logout__roles__=None
    def logout(self, REQUEST):
        """Logout"""
        REQUEST['RESPONSE'].expireCookie('__ac', path='/')
        return self.docLogout(self, REQUEST)
    




manage_addUserLDAPForm=HTMLFile('addUserLDAP', globals())

def manage_addUserLDAP(self, connection_id, scope, 
		       user_name_attr, password_attr,
		       roles_attr, domains_attr,
		       cookie_mode=0, REQUEST=None):
    """ """
    if hasattr(self.aq_base, 'acl_users'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a User Folder',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])
    
    basedn =  getattr(self, connection_id).dn
    
    ob=UserLDAP(connection_id, scope, basedn, 
		user_name_attr, password_attr, 
		roles_attr, domains_attr,
		cookie_mode)

    self._setObject('acl_users', ob)
    self.__allow_groups__=self.acl_users
    if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)





addr_match=regex.compile('[0-9\.\*]*').match
host_match=regex.compile('[A-Za-z0-9\.\*]*').match


def domainSpecValidate(spec):
    for ob in spec:
        sz=len(ob)
        if not ((addr_match(ob) == sz) or (host_match(ob) == sz)):
            return 0
    return 1


def domainSpecMatch(spec, request):

    if request.has_key('REMOTE_HOST'):
        host=request['REMOTE_HOST']
    else: host=''

    if request.has_key('REMOTE_ADDR'):
        addr=request['REMOTE_ADDR']
    else: addr=''

    if not host and not addr:
        return 0

    if not host:
        host=socket.gethostbyaddr(addr)[0]
    if not addr:
        addr=socket.gethostbyname(host)

    _host=split(host, '.')
    _addr=split(addr, '.')
    _hlen=len(_host)
    _alen=len(_addr)
    
    for ob in spec:
        sz=len(ob)
        _ob=split(ob, '.')
        _sz=len(_ob)

        if addr_match(ob)==sz:
            if _sz != _alen:
                continue
            fail=0
            for i in range(_sz):
                a=_addr[i]
                o=_ob[i]
                if (o != a) and (o != '*'):
                    fail=1
                    break
            if fail:
                continue
            return 1

        if host_match(ob)==sz:
            if _hlen < _sz:
                continue
            elif _hlen > _sz:
                _item=_host[-_sz:]
            else:
                _item=_host
            fail=0
            for i in range(_sz):
                h=_item[i]
                o=_ob[i]
                if (o != h) and (o != '*'):
                    fail=1
                    break
            if fail:
                continue
            return 1
    return 0


# Check for the necessity of this
# Do we have to worry about case in LDAP

def ldapattr(ob, attr):
    name=attr
    if hasattr(ob, attr):
        return str(getattr(ob,attr))
    attr=upper(attr)
    if hasattr(ob, attr):
        return str(getattr(ob,attr))
    attr=lower(attr)
    if hasattr(ob, attr):
        return str(getattr(ob,attr))
    raise NameError, name

def absattr(attr):
    if callable(attr): return attr()
    return attr

def reqattr(request, attr):
    try:    return request[attr]
    except: return None


_docLogin="""<HTML>
<HEAD>
<TITLE>Login</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">
<P>
<FORM ACTION="<!--#var SCRIPT_NAME--><!--#var PATH_INFO--><!--#if QUERY_STRING-->?<!--#var QUERY_STRING--><!--#/if-->" METHOD="POST">
<TABLE>
<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <STRONG>Name</STRONG>
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="TEXT" NAME="__ac_name" SIZE="20">
  </TD>
</TR>

<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <STRONG>Password</STRONG>
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="PASSWORD" NAME="__ac_password" SIZE="20">
  </TD>
</TR>
<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="SUBMIT" NAME="submit" VALUE=" Ok ">
  </TD>
</TR>
</TABLE>
</FORM>
</BODY>
</HTML>"""


_docLogout="""<HTML>
<HEAD>
<TITLE>Logout</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">
<P>
<CENTER>
You have been logged out of the system.
</CENTER>
</BODY>
</HTML>"""


